/*
 Functions to discover simple blocks of executable code in NAND flash, load and execute them
 
 Matt Evans, 8 May 2005
 */

#include "asmfns.h"
#include "pblcalls.h"

#include "boot.h"


inline int validBootableBlock(struct bootBlockHeader *bbh)
{
  return (bbh->magic1 == BOOT_MAGIC1) && (bbh->magic2 == BOOT_MAGIC2);
}

// Returns block number of first executable image, or <0 on error.
// Errors: -1 not found (... -2 interrupted (not impl))
int findFirstBootableBlock(void)
{
  // Iterate through calls to findBootableBlock & list what we find.
  unsigned char blockBuffer[8192];
  int found;
  
  // Start at the block after the monitor's length; PBL header's 0x7c, then 
  // we know how much space we use.  Round up to next 8K page:
  int block = ((monitorLength + 0x7c + 8191) & ~0x1fff) >> 13;
  
  found = findBootableBlock(block, blockBuffer);
  
  if (found != -1)
  {
    struct bootBlockHeader *bbh = (struct bootBlockHeader *)blockBuffer;
    uart_putstring("Found bootable image at flash address ");
    uart_phex32(found*8192);
    uart_putstring(", length ");
    uart_phex32(bbh->length);
    uart_putstring(" (\"");
    uart_putstring((char*)(blockBuffer + bbh->offsetToDescr));
    uart_putstring("\")\r\n");
  }    
  return found;
}

// Passed a block number index into flash, search upwards and check each block to see if it
// has a bootable header on it.  If so, return its block index, and fill in the 8K buffer passed
// with its contents.  If nothing is found (i.e. it gets to the end), returns -1.  Buffer is 
// always corrupted...  Can be used to list bootable blocks, call & it returns the first.  
// Call with the previous block num+1, and it returns the next, etc.
int findBootableBlock(int from, unsigned char *blockSizeBuffer)
{
  while (from < (MAXFLASHADDR/8192))
  {
    if (pbl_readflash(blockSizeBuffer, from*8192, 8192) != 1)
    {
      // stop - nothing found ...
      return -1;
    }
    
    if (validBootableBlock((struct bootBlockHeader *)blockSizeBuffer))
    {
      return from;
    }
    else
    {
      from++;
    }
  }
  // Didn't find one, got to the end of the flash.
  return -1;
}

void listBootableBlocks(void)
{
  // Iterate through calls to findBootableBlock & list what we find.
  unsigned char blockBuffer[8192];
  
  // Start at the block after the monitor's length; PBL header's 0x7c, then 
  // we know how much space we use.  Round up to next 8K page:
  int block = ((monitorLength + 0x7c + 8191) & ~0x1fff) >> 13;
  
  uart_putstring("Bootable blocks:\r\n");
  do
  {
    block = findBootableBlock(block, blockBuffer);
    if (block != -1)
    {
      struct bootBlockHeader *bbh = (struct bootBlockHeader *)blockBuffer;
      uart_putstring("Found bootable image at flash address ");
      uart_phex32(block*8192);
      uart_putstring(", length ");
      uart_phex32(bbh->length);
      uart_putstring(" (\"");
      uart_putstring((char*)(blockBuffer + bbh->offsetToDescr));
      uart_putstring("\")\r\n");
      
      // Start searching for the next one after the end of this bootable image:
      block += ((bbh->length + 8191) & ~0x1fff) >> 13;  // round its length up a block
    }
  } while (block != -1);
}

// Boot (or try to) an image.  Pass in a block number (i.e. flash address/8192)
void bootThatImage(int imageblock) 
{ 
  unsigned char blockSizeBuffer[8192];
  struct bootBlockHeader *bbh = (struct bootBlockHeader *)blockSizeBuffer;

  // Inept inefficient version: read the first block twice.  Once to get
  // the header, a second time to actually load it.
  
  if (imageblock < (MAXFLASHADDR/8192))
  {
    if (pbl_readflash(blockSizeBuffer, imageblock*8192, 8192) != 1)
    {
      uart_putstring("bootThatImage: pbl_readflash failed!\r\n");
      return;
    }
    if (!validBootableBlock((struct bootBlockHeader *)blockSizeBuffer))
    {
      uart_putstring("bootThatImage: Block ");
      uart_phex32(imageblock);
      uart_putstring(" not valid bootable block\r\n");
      return;  // FIXME, complain more loudly
    }
    
    // Don't let us load anything a) over the monitor, b) over PBL, c) over the stacks etc.
    if (bbh->loadAddress > (unsigned int)&_end)
    {
      // All good.  Load it:
      
      uart_putstring("Loading image \"");
      uart_putstring((char*)(blockSizeBuffer + bbh->offsetToDescr));
      uart_putstring("\" to ");
      uart_phex32(bbh->loadAddress);
      if (pbl_readflash((unsigned char *)bbh->loadAddress, imageblock*8192, (bbh->length + 8191) & ~8191) != 1)
      {
        uart_putstring(" - argh! failed!\r\n");
        return;
      }      
      else
      {
        uart_putstring(" - done\r\nJumping to entry address ");
        uart_phex32(bbh->entryAddress);
        uart_putstring(":\r\n");
        
        // Jump.  goodbye!
        asm __volatile__ ("mov r14, pc ; mov pc, %0" : : "r"(bbh->entryAddress) : "r14");
      }
    }
    else
    {
      uart_putstring("Image load address ");
      uart_phex32(bbh->loadAddress);
      uart_putstring(" overlaps PBL/Monitor (below ");
      uart_phex32(_end);
      uart_putstring("), not loading.\r\n");
    }   
  }
}
