/*
 cons.c
 
 Simple debug output on LCD:  Provides printstring and printhex functions and a full
 ASCII set.  The only cursor movement so far supported is cls and \n ;-)
 
 (c) Matt Evans 25 Apr 05
 */

#include "cons.h"
#include "asmfns.h"
#include "libcfns.h"
#include "peripherals.h"

volatile unsigned char *screen_mem = (volatile char *)0x20000;
int cur_x;
int cur_y;

extern unsigned char *fontdata_8x8;

volatile unsigned int *lcdc_base = (volatile unsigned int *)LCDC_BASE;
volatile unsigned int *lcdc_pal = (volatile unsigned int *)LCDC_PAL;
volatile unsigned int *lcdc_base_ext = (volatile unsigned int *)LCDC_EXT_BASE;
volatile unsigned int *pwm_base = (volatile unsigned int *)PWM_BASE;

/* Set up the LCDC
   This is modelled on the way that the usual firmware splashscreen does it, since 
   I know that doesn't explode it.

   It'd be nice to be able to store our preferred contrast somewhere.
*/
void cons_init(void)
{
  int i;
  
  // power off the display first!
  lcdc_base[7] = 0;  // All off
  
  lcdc_base_ext[0] = 0x000c;
  lcdc_base_ext[2] = 0;
  lcdc_base_ext[3] = 0;
  lcdc_base_ext[1] = 0x28;  // Backlight on (bit 0x20)

  lcdc_base[0] = 0x274; // 0x295.. yeh
  lcdc_base[1] = 0x13f;
  lcdc_base[2] = 0x007702ca;
  lcdc_base[3] = 0;
  // The framebuffer runs from 0x20000 to 0x32c00:
  lcdc_base[4] = 0x00020000; 
  lcdc_base[5] = 0x00020000;
  lcdc_base[6] = 0;
    
  // Set up palette:
  // I THINK only the first 16 entries are necessary:
  for (i = 0; i < 16; i += 2)
  {
    lcdc_pal[i/2] = (i << 1) | ((i+1) << 17);
  }
  // When reading palette from, say, splashscreen make sure to disable LCDC as reads are disabled when it runs.

  // Set up the PWM driver (no. 2) to provide the contrast voltage to the LCD:
  pwm_base[PWM0_ENB/4] = 0;
  pwm_base[PWM1_ENB/4] = 0;
  pwm_base[PWM2_ENB/4] = 0;
  pwm_base[PWM3_ENB/4] = 0;
  // PWM2:
  pwm_base[PWM2_TC/4] = 0x61a8;
  pwm_base[PWM2_DC/4] = 0x3200; // This is the contrast, ranges from 0xc4 to 0x613c (FIXME, have a set fn!)
  pwm_base[PWM2_DIV/4] = 2;
  pwm_base[PWM2_SYNC/4] = 0;
  pwm_base[PWM2_INV/4] = 0;
  pwm_base[PWM2_ENB/4] = 1; // ...and enable it.

  // Finally the ctrl reg
  lcdc_base[7] = 0x415 | 0x10000;   // LCD controller start
  usleep(20000);
  lcdc_base[7] = 0xc15 | 0x10000;   // LCD power up (815 = LE pix, c15 = BE), lower bandwidth FIFO watermark
                          // Little endian, and little endian pixel packing
                          // Oh don't ask, what a ballache - the pixel order of this thing is awful.
                          // See some of the convoluted stuff below to draw stuff...
  
  cons_cls();
}

void cons_cls(void)
{
  asm("mov r0,#0 \n"
      "mov r1,#0 \n"
      "mov r2,#0 \n"
      "mov r3,#0 \n"
      "mov r4,#0 \n"
      "mov r5,#0 \n"
      "mov r6,#0 \n"
      "mov r7,#0 \n"
      "mov r8,#2400 \n"
      "1: "
      "stmia %0!, {r0-r7} \n"
      "subs r8,r8,#1 \n"
      "bgt 1b \n"
      : : "r"(screen_mem) 
      : "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7", "r8");
  cur_x = 0;
  cur_y = 0;
}

void cons_putbyte(char c)
{
  volatile unsigned char *f = screen_mem + (cur_x * 4) + (cur_y*4*CONS_WIDTH*8);

  // Is it a non-printingy newline?
  if (c == '\n')
  {
    // Clear from current x to the end of the line
    asm("mov   r0, #0 \n"
        "mov   r1, %0 \n"  // the screen address to start at
        "mov   r4, #8 \n"
        "2: \n"
        "mov   r3, r1 \n"
        "add   r1, r1, #480/2 \n"
        "mov   r2, %1 \n"  // the number of words per row
        "1: \n"
        "str   r0, [r3], #4 \n"
        "subs  r2, r2, #1 \n"
        "bgt   1b \n"
        "subs  r4, r4, #1 \n"
        "bgt   2b \n" : : "r"(f), "r"(CONS_WIDTH - cur_x) : "r0", "r1", "r2", "r3", "r4");
    
    cur_x = 0;
    if (++cur_y >= CONS_HEIGHT)
    {
      cur_y = 0;
    }
  }
  else
  {
    if (++cur_x >= CONS_WIDTH)
    {
      cur_x = 0;
      if (++cur_y >= CONS_HEIGHT)
      {
        cur_y = 0;
      }
    }
    
    /* The display is in some stupid halfbyte-bigendian order!
    */
    asm("add   %2, %1, %2, lsl#3 \n"
        "mov   r2, #8 \n"
        "1: \n"
        "ldrb  r0, [%2], #1\n"
        "mov   r1, #0 \n"
        // original amstrad order
        "tst   r0, #0x01 \n"
        "orrne r1, r1, #0x00f00000 \n" // right
        "tst   r0, #0x02 \n"
        "orrne r1, r1, #0x000f0000 \n" // 2nd right
        "tst   r0, #0x04 \n"
        "orrne r1, r1, #0xf0000000 \n" // 3rd right
        "tst   r0, #0x08 \n"      
        "orrne r1, r1, #0x0f000000 \n" // 4th right/middle
        "tst   r0, #0x10 \n"
        "orrne r1, r1, #0x000000f0 \n" // 4th left/middle
        "tst   r0, #0x20 \n"
        "orrne r1, r1, #0x0000000f \n" // 3rd left      
        "tst   r0, #0x40 \n"
        "orrne r1, r1, #0x0000f000 \n" // 2nd left
        "tst   r0, #0x80 \n"
        "orrne r1, r1, #0x00000f00 \n" // left
        
        /*   // LE byte order (supposedly), le pixel order: (no still wrong)
        "tst   r0, #0x01 \n"
        "orrne r1, r1, #0x00f00000 \n"
        "tst   r0, #0x02 \n"
        "orrne r1, r1, #0x000f0000 \n"
        "tst   r0, #0x04 \n"
        "orrne r1, r1, #0xf0000000 \n"
        "tst   r0, #0x08 \n"      
        "orrne r1, r1, #0x0f000000 \n"
        "tst   r0, #0x10 \n"
        "orrne r1, r1, #0x000000f0 \n"
        "tst   r0, #0x20 \n"
        "orrne r1, r1, #0x0000000f \n"
        "tst   r0, #0x40 \n"
        "orrne r1, r1, #0x0000f000 \n"
        "tst   r0, #0x80 \n"
        "orrne r1, r1, #0x00000f00 \n"
        */
        "str   r1, [%0], #240 \n"
        "subs  r2, r2, #1 \n"
        "bgt   1b \n"
        : : "r"(f), "r"(&fontdata_8x8), "r"(c)
        : "r0", "r1", "r2");
  }
}

void cons_putstring(char *string)
{
  char c;
  while ((c = *string++) != 0)
  {
    cons_putbyte(c);
  }
}

void cons_phex8(int n)
{
  int i;
  n = n << 24;
  for (i = 0; i < 2; i++)
  {
    int a = (n >> 28) & 0xf;
    cons_putbyte( (a >= 10) ? (a + ('a' - 10)) : (a + '0') );
    n = n << 4;
  }
}

void cons_phex32(int n)
{
  int i;

  for (i = 0; i < 8; i++)
  {
    int a = (n >> 28) & 0xf;
    cons_putbyte( (a >= 10) ? (a + ('a' - 10)) : (a + '0') );
    n = n << 4;
  }
}

inline int pixelToBlah(int p)
{
  // Given a left-to-right pixel in a word (there are 8,
  // nybble-sized) return the actual location of the nybble
  // in the word.
  // This is just p ^ 2 !
  return p^2;
}

unsigned char *firePixels = (unsigned char *)0x700000;
#define FIRE_CTRL_POINTS 11
void cons_fire(void)
{
  int x, y, p;
  int fireCtrlPoints[FIRE_CTRL_POINTS];  // every 60 pixels
  char k = 0;
  
  do
  {
    /* Random pixels on the bottom row:
       But not quite random; I'd like strong areas blending into each other.
    */
    for (p = 0; p < FIRE_CTRL_POINTS; p++)
    {
      fireCtrlPoints[p] += (((int)random()) >> 27);  // goes up & down
      fireCtrlPoints[p] &= 0xff;
    }
    // Now interpolate between those ctrl points:
    x = 0;
    for (p = 0; p < FIRE_CTRL_POINTS-1; p++)
    {
      int current = fireCtrlPoints[p];
      int next = fireCtrlPoints[p+1];
      int diff1616 = ((next - current)<<16)/(480/(FIRE_CTRL_POINTS-1));
      int blendpix;
      int current1616 = current<<16;
      for (blendpix = 0; blendpix < (480/(FIRE_CTRL_POINTS-1)); blendpix++)
      {
        current1616 += diff1616;
        firePixels[(319*480) + x++] = (current1616>>16);
      }
    }
    
    // blur up:
    for (y = 318; y >= 0; y--)
    {
      for (x = 0; x < 480; x++)
      {
        unsigned char l = (x > 0) ? firePixels[(y*480) + (x-1)] : 0;
        unsigned char r = (x < 479) ? firePixels[(y*480) + (x+1)] : 0;
//        unsigned char d = (y < 319) ? firePixels[((y+1)*480) + x] : 0;
        unsigned char dr = ((x < 479) && (y < 319)) ? firePixels[((y+1)*480) + (x+1)] : 0;
        unsigned char dl = ((x > 0) && (y < 319)) ? firePixels[((y+1)*480) + (x-1)] : 0;
        
//        firePixels[(y*480) + x] = (l+r+d+dr+dl)/5;
        firePixels[(y*480) + x] = (l+r+dr+dl)/4;
      }
    }
    
    // Splat to screen:
    for (y = 0; y < 320; y++)
    {
      for (x = 0; x < 480; x += 8)
      {
        unsigned int wordOfPixels = 0;
        // Read 8 pixels from buffer, pack into u32 word & write to displ buffer:
        unsigned int bot = *((unsigned int *)&firePixels[(y*480)+x]);
        unsigned int top = *((unsigned int *)&firePixels[(y*480)+x+4]);
        // Pack these 32bit words into 16bits/4nybbles: (masking out the top nybble from each byte-wide pixel above)
        wordOfPixels = (bot << 4)   & 0x00000f00;  // 0 0 0 0 0 n0 0 0 
        wordOfPixels |= bot         & 0x0000f000;  // 0 0 0 0 n1 n0 0 0 
        wordOfPixels |= (bot >> 20) & 0x0000000f;  // 0 0 0 0 n1 n0 0 n2 
        wordOfPixels |= (bot >> 24) & 0x000000f0;  // 0 0 0 0 n1 n0 n3 n2 
      
        wordOfPixels |= (top << 20) & 0x0f000000;  // 0 n4 0 0 n1 n0 n3 n2
        wordOfPixels |= (top << 16) & 0xf0000000;  // n5 n4 0 0 n1 n0 n3 n2
        wordOfPixels |= (top >> 4)  & 0x000f0000;  // n5 n4 0 n6 n1 n0 n3 n2
        wordOfPixels |= (top >> 8)  & 0x00f00000;  // n5 n4 n7 n6 n1 n0 n3 n2
        
        ((unsigned int *)screen_mem)[(y*60) + (x>>3)] = wordOfPixels;
      }
    }
  }
  while (!uart_testrx() || ((k = uart_getbyte()) != 'q'));
}
